package org.processmining.plugins.complianceframework.model;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;

import weka.classifiers.trees.J48;
import weka.core.Attribute;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;

public class InstanceClassifier implements Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = -2599421158533460476L;
	private static final String UNDEFINED_VALUE = "#undefined#";
	public static final String[] timePrefixes = new String[] { "Time between start of the case and the classified activity",
			"Time between the classified activity and the one previous to it",
			"Time between the classified activity and the selected one" };

	//	private static final String[] classAttributeVals = new String[] { "log-only", "model-only", "incorrect data",
	//			"incorrect time", "log-only with incorrect time", "model-only with incorrect time",
	//			"incorrect data and time", "non-problematic" };
	private List<String> classAttributeValsList;

	private List<String> activityClassificationClassValues = null;
	public int activityClassificationInstanceCount = 0;
	private List<String> assignmentClassificationClassValues = null;
	public int assignmentClassificationInstanceCount = 0;
	private List<String> intersectionClassificationClassValues = null;
	public int intersectionClassificationInstanceCount = 0;

	private Map<String, List<String>> activityAttributePossibleVals;
	private Map<String, List<String>> assignmentAttributePossibleVals;
	private Map<String, List<String>> intersectionAttributePossibleVals;
	//Time Attribute modes: 0 - disable, 1 - time between start of trace and the problematic activity, 2 - between previous activity and the problematic activity, 3 - between chosen(anchor) activity and the problematic activity
	public int activityTimeAttributeMode;
	public int assignmentTimeAttributeMode;
	public int intersectionTimeAttributeMode;
	private String activityAnchor;
	private String assignmentAnchor;
	private String intersectionAnchor;

	// private List<String> activities;
	private Map<String, float[]> intervals;
	private J48 activityTree = null;
	private J48 assignmentTree = null;
	private J48 intersectionTree = null;

	private Instances activityInstances = null;
	private Instances assignmentInstances = null;
	private Instances intersectionInstances = null;

	//	private PNRepResult controFlowAlignment = null;
	private DashboardResultReplay replayResultWithData = null;
	private DashboardResultReplay replayResultTemporal = null;
	private ComplianceDashboard dashboard;
	public ProblemSentenceObject problemSentenceObject;

	private List<String> attributeIndex;
	private List<String> activityClassificationAttributes;
	private List<String> assignmentClassificationAttributes;
	private List<String> intersectionClassificationAttributes;

	private XLog log;

	// TODO: do we need control flow alignment?
	// the value with index 0 in attributePossibleVals must be empty String for
	// each attribute
	public InstanceClassifier(ComplianceDashboard dashboard, ProblemSentenceObject problemSentenceObject,
			Map<String, float[]> intervals, XLog log) {
		init(dashboard, problemSentenceObject, intervals, log, 0, 0, 0, null, null, null);
	}

	public InstanceClassifier(ComplianceDashboard dashboard, ProblemSentenceObject problemSentenceObject,
			Map<String, float[]> intervals, XLog log, int activityTimeAttributeMode, int assignmentTimeAttributeMode,
			int intersectionTimeAttributeMode, String activityAnchor, String assignmentAnchor, String intersectionAnchor) {
		init(dashboard, problemSentenceObject, intervals, log, activityTimeAttributeMode, assignmentTimeAttributeMode,
				intersectionTimeAttributeMode, activityAnchor, assignmentAnchor, intersectionAnchor);
	}

	public void init(ComplianceDashboard dashboard, ProblemSentenceObject problemSentenceObject,
			Map<String, float[]> intervals, XLog log, int activityTimeAttributeMode, int assignmentTimeAttributeMode,
			int intersectionTimeAttributeMode, String activityAnchor, String assignmentAnchor, String intersectionAnchor) {
		this.activityAttributePossibleVals = new HashMap<String, List<String>>();
		this.assignmentAttributePossibleVals = new HashMap<String, List<String>>();
		this.intersectionAttributePossibleVals = new HashMap<String, List<String>>();
		this.attributeIndex = new ArrayList<String>();
		this.activityAttributePossibleVals = new HashMap<String, List<String>>();
		this.assignmentAttributePossibleVals = new HashMap<String, List<String>>();
		this.intersectionAttributePossibleVals = new HashMap<String, List<String>>();
		for (String attribute : dashboard.getAttributeMovesMap().keySet()) {
			if (!intervals.containsKey(attribute)) {
				this.activityAttributePossibleVals.put(attribute, new ArrayList<String>());
				this.assignmentAttributePossibleVals.put(attribute, new ArrayList<String>());
				this.intersectionAttributePossibleVals.put(attribute, new ArrayList<String>());
			}
			this.attributeIndex.add(attribute);
			//			for (String attrValue : dashboard.getAttributeMovesMap().get(attribute).keySet()) {
			//				this.attributePossibleVals.get(attribute).add(attrValue);
			//			}
		}

		this.problemSentenceObject = problemSentenceObject;
		//		this.controFlowAlignment = dashboard.getControFlowAlignment();
		this.replayResultWithData = dashboard.getReplayResultWithData();
		this.replayResultTemporal = dashboard.getReplayResultTemporal();
		this.dashboard = dashboard;
		this.intervals = intervals;
		this.log = log;

		this.activityClassificationInstanceCount = obtainInstanceCount(0);
		this.assignmentClassificationInstanceCount = obtainInstanceCount(1);
		this.intersectionClassificationInstanceCount = obtainInstanceCount(2);

		this.activityTimeAttributeMode = activityTimeAttributeMode;
		this.assignmentTimeAttributeMode = assignmentTimeAttributeMode;
		this.intersectionTimeAttributeMode = intersectionTimeAttributeMode;
		this.activityAnchor = activityAnchor;
		this.assignmentAnchor = assignmentAnchor;
		this.intersectionAnchor = intersectionAnchor;
	}

	public int obtainInstanceCount(int classificationType) {
		int result = 0;
		Map<String, List<String>> attributePossibleVals = null;

		if (classificationType == 0) {
			attributePossibleVals = this.activityAttributePossibleVals;
		} else if (classificationType == 1) {
			attributePossibleVals = this.assignmentAttributePossibleVals;
		} else if (classificationType == 2) {
			attributePossibleVals = this.intersectionAttributePossibleVals;
		}

		for (DashboardAlignment dataAlignment : this.replayResultWithData.labelStepArray) {
			for (int i = 0; i < dataAlignment.getLogTrace().size(); i++) {
				
				boolean isDecisionPoint = dataAlignment.getProcessTrace().get(i).keySet().size() > 0
						&& dataAlignment.getProcessTrace().get(i).keySet().size() > 0;
				ExecutionStep currentStep;
				String currentActivity;
				if (dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MINVI && !isDecisionPoint)
					//						|| (dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MREAL && classificationType != 0))
					continue;
				if (dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MREAL
						|| dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MINVI) {
					currentStep = dataAlignment.getLogTrace().get(i);
					currentActivity = dataAlignment.getProcessTrace().get(i).getActivity();
				} else {
					currentStep = dataAlignment.getLogTrace().get(i);
					currentActivity = dashboard.getActivityName(currentStep.getActivity());
				}

				if (!includeStepInInstances(currentStep, currentActivity, classificationType))
					continue;
				for (Entry<String, Object> attributeEntry : currentStep.entrySet()) {
					if (!this.intervals.containsKey(attributeEntry.getKey()) && attributeEntry.getValue() != null
							&& !attributePossibleVals.get(attributeEntry.getKey()).contains(attributeEntry.getValue())) {
						attributePossibleVals.get(attributeEntry.getKey()).add(attributeEntry.getValue().toString());
					}
				}
				result++;
			}
		}
		for (Entry<String, List<String>> attributeEntry : attributePossibleVals.entrySet()) {
			if (!this.intervals.containsKey(attributeEntry.getKey()) && attributeEntry.getValue() != null) {
				attributeEntry.getValue().add(UNDEFINED_VALUE);
			}
		}
		return result;
	}

	public void generateTrees(String[] activityOptions, String[] assignmentOptions, String[] intersectionOptions) {
		Instances instances;
		if (this.problemSentenceObject.getActivity() != null && this.problemSentenceObject.getActivity().length() > 0
				&& (this.activityClassificationAttributes.size() > 0 || this.activityTimeAttributeMode != 0)) {
			this.configureClassAttributeVals(0);
			this.activityClassificationClassValues = this.classAttributeValsList;
			instances = this.createInstances(this.activityClassificationAttributes, 0);
			this.activityInstances = instances;
			this.activityTree = this.generateTree(instances, activityOptions);
		}
		if (this.problemSentenceObject.getAttribute() != null && this.problemSentenceObject.getAttribute().length() > 0
				&& this.problemSentenceObject.getAttributeValue() != null
				&& this.problemSentenceObject.getAttributeValue().length() > 0
				&& (this.assignmentClassificationAttributes.size() > 0 || this.assignmentTimeAttributeMode != 0)) {
			this.configureClassAttributeVals(1);
			this.assignmentClassificationClassValues = this.classAttributeValsList;
			instances = this.createInstances(this.assignmentClassificationAttributes, 1);
			this.assignmentInstances = instances;
			this.assignmentTree = this.generateTree(instances, assignmentOptions);
		}
		if (this.problemSentenceObject.getActivity() != null && this.problemSentenceObject.getActivity().length() > 0
				&& this.problemSentenceObject.getAttribute() != null
				&& this.problemSentenceObject.getAttribute().length() > 0
				&& this.problemSentenceObject.getAttributeValue() != null
				&& this.problemSentenceObject.getAttributeValue().length() > 0
				&& (this.intersectionClassificationAttributes.size() > 0 || this.intersectionTimeAttributeMode != 0)) {
			this.configureClassAttributeVals(2);
			this.intersectionClassificationClassValues = this.classAttributeValsList;
			instances = this.createInstances(this.intersectionClassificationAttributes, 2);
			this.intersectionInstances = instances;
			this.intersectionTree = this.generateTree(instances, intersectionOptions);
		}
		//		
		//		Instances instances = this.createInctances(this.attributeClassificationAttributes, 2);
		//		this.attributeTree = this.generateTree(instances);
	}

	// 1 instance == 1 event
	// instance attributes: event name; values of all event attributes except
	// timestamp; meets temporal conditions or not; class attribute: violating
	// or not
	// classificationType: 0 - activity classification; 1 - assignment classification; 2 - attribute classification
	private Instances createInstances(List<String> attributesToIncludeList, int classificationType) {
		ArrayList<Attribute> includedAttributesList = new ArrayList<Attribute>();
		Map<String, List<String>> attributePossibleVals = null;
		Map<ExecutionStep, Float> timeClassificationValues = null;

		Attribute classAttribute = new Attribute("Problem type", this.classAttributeValsList);
		int timeAttributeMode = 0;
		String anchorActivity = null;
		if (classificationType == 0) {
			timeAttributeMode = this.activityTimeAttributeMode;
			anchorActivity = this.activityAnchor;
			attributePossibleVals = this.activityAttributePossibleVals;
		} else if (classificationType == 1) {
			timeAttributeMode = this.assignmentTimeAttributeMode;
			anchorActivity = this.assignmentAnchor;
			attributePossibleVals = this.assignmentAttributePossibleVals;
		} else if (classificationType == 2) {
			timeAttributeMode = this.intersectionTimeAttributeMode;
			anchorActivity = this.intersectionAnchor;
			attributePossibleVals = this.intersectionAttributePossibleVals;
		}

		List<String> activitiesList = null;
		if (attributesToIncludeList.contains("Activity")) {
			activitiesList = new ArrayList<String>();
			for (String activity : this.dashboard.getActivityMovesMap().keySet()) {
				int[] activityStats = this.dashboard.getActivityMovesMap().get(activity);
				int totalActivityMoves = activityStats[0] + activityStats[1] + activityStats[activityStats.length - 2];
				//only add activity if it has visible moves
				if (totalActivityMoves > activityStats[0] && !activitiesList.contains(activity))
					activitiesList.add(activity);
			}
			includedAttributesList.add(new Attribute("Activity", activitiesList));
		}

		for (String attrName : attributesToIncludeList) {
			if (attrName.equals("Activity"))
				continue;
			if (!intervals.containsKey(attrName)) {
				List<String> possibleVals = new ArrayList<String>(attributePossibleVals.get(attrName));
				includedAttributesList.add(new Attribute(attrName, possibleVals));
			}
			else
				includedAttributesList.add(new Attribute(attrName));
		}

		if (timeAttributeMode != 0) {
			includedAttributesList.add(new Attribute("timeClass"));
			//		}

			//		if (timeAttributeMode == 3) {
			timeClassificationValues = new HashMap<ExecutionStep, Float>();
			long anchorTimestamp, newAnchorTimestamp, latestRealTimestamp;

			for (DashboardAlignment dataAlignment : this.replayResultWithData.labelStepArray) {
				anchorTimestamp = -1;
				newAnchorTimestamp = -1;
				latestRealTimestamp = -1;
				for (int i = 0, l = -1; i < dataAlignment.getLogTrace().size(); i++) {

					if (i == 0) {
						latestRealTimestamp = getTimestampAttributeValue(dataAlignment.getTraceName(), 0);
						if (timeAttributeMode == 1) {
							//							anchorTimestamp = latestRealTimestamp;
							newAnchorTimestamp = latestRealTimestamp;
						}
					}
					ExecutionStep currentStep;
					String currentActivity;

					currentStep = dataAlignment.getLogTrace().get(i);
					if (dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MREAL
							|| dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MINVI) {

						currentActivity = dataAlignment.getProcessTrace().get(i).getActivity();
					} else {
						l++;
						currentActivity = dashboard.getActivityName(currentStep.getActivity());

						latestRealTimestamp = getTimestampAttributeValue(dataAlignment.getTraceName(), l);
						if (timeAttributeMode == 2)
							newAnchorTimestamp = latestRealTimestamp;
						if (timeAttributeMode == 3 && currentActivity.equals(anchorActivity))
							newAnchorTimestamp = latestRealTimestamp;
					}

					if (!includeStepInInstances(currentStep, currentActivity, classificationType)) {
						anchorTimestamp = newAnchorTimestamp;
					} else {
						if (timeAttributeMode == 3 && anchorTimestamp != -1) {
							if (dataAlignment.getStepTypes().get(i) != DataAwareStepTypes.MREAL
									|| dataAlignment.getStepTypes().get(i) != DataAwareStepTypes.MINVI) {
								timeClassificationValues.put(currentStep,
										getTimeClassificationValue(dataAlignment.getTraceName(), l, anchorTimestamp));
							} else
								timeClassificationValues.put(currentStep,
										getTimeClassificationValue(anchorTimestamp, latestRealTimestamp));
						} else {
							String subCurrentActivity;
							ExecutionStep subCurrentStep;
							long problemTimestamp = latestRealTimestamp;
							for (int j = i + 1, k = l; j < dataAlignment.getLogTrace().size(); j++) {
								subCurrentStep = dataAlignment.getLogTrace().get(j);
								if (dataAlignment.getStepTypes().get(j) == DataAwareStepTypes.MREAL
										|| dataAlignment.getStepTypes().get(j) == DataAwareStepTypes.MINVI) {

									subCurrentActivity = dataAlignment.getProcessTrace().get(j).getActivity();
								} else {
									k++;
									subCurrentActivity = dashboard.getActivityName(subCurrentStep.getActivity());

									latestRealTimestamp = getTimestampAttributeValue(dataAlignment.getTraceName(), k);
								}

								if (subCurrentActivity.equals(anchorActivity)) {
									newAnchorTimestamp = latestRealTimestamp;
									break;
								}
							}
							timeClassificationValues.put(currentStep,
									getTimeClassificationValue(newAnchorTimestamp, problemTimestamp));
						}
						anchorTimestamp = newAnchorTimestamp;
					}
				}
			}
		}

		includedAttributesList.add(classAttribute);
		Instances instances = new Instances("StepInstances", includedAttributesList, 0);
		instances.setClassIndex(includedAttributesList.size() - 1);
		//		long anchorTimestamp, newAnchorTimestamp;
		//		long latestRealTimestamp;
		for (DashboardAlignment dataAlignment : this.replayResultWithData.labelStepArray) {
			//			anchorTimestamp = -1;
			//			newAnchorTimestamp = -1;
			//			latestRealTimestamp = -1;
			for (int i = 0, l = -1; i < dataAlignment.getLogTrace().size(); i++) {
				try {
					//if process step and log step both contain attributes, then the step is a decision point
					boolean isDecisionPoint = dataAlignment.getProcessTrace().get(i).keySet().size() > 0
							&& dataAlignment.getProcessTrace().get(i).keySet().size() > 0;
					//					if (classificationType == 0)
					//					if (i == 0) {
					//						latestRealTimestamp = getTimestampAttributeValue(dataAlignment.getTraceName(), 0);
					//						if (timeAttributeMode == 1) {
					//							//							anchorTimestamp = latestRealTimestamp;
					//							newAnchorTimestamp = latestRealTimestamp;
					//						}
					//					}

					if (dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MINVI && !isDecisionPoint)
						//							|| (dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MREAL && classificationType != 0))
						continue;
					ExecutionStep currentStep;
					String currentActivity;
					currentStep = dataAlignment.getLogTrace().get(i);
					if (dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MREAL
							|| dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MINVI) {
						currentActivity = dataAlignment.getProcessTrace().get(i).getActivity();
					} else {
						l++;
						currentActivity = dashboard.getActivityName(currentStep.getActivity());

						//						latestRealTimestamp = getTimestampAttributeValue(dataAlignment.getTraceName(), l);
						//												if (timeAttributeMode == 2)
						//													newAnchorTimestamp = latestRealTimestamp;
						//												if (timeAttributeMode == 3 && currentActivity.equals(anchorActivity))
						//													newAnchorTimestamp = latestRealTimestamp;

					}
					if (!includeStepInInstances(currentStep, currentActivity, classificationType)) {
						//						anchorTimestamp = newAnchorTimestamp;
						continue;
					}

					double[] instanceValue = new double[instances.numAttributes()];
					int j = -1;
					for (String attribute : attributesToIncludeList) {
						j++;
						if (attribute.equals("Activity")) {
							instanceValue[0] = activitiesList.indexOf(currentActivity);
						} else {
							//							for (int j = 1; j < instanceValue.length - 1; j++) {

							if (currentStep.containsKey(attribute) && currentStep.get(attribute) != null) {
								if (intervals.containsKey(attribute)) {
									if (currentStep.get(attribute) instanceof Number) {
										instanceValue[j] = ((Number) currentStep.get(attribute)).doubleValue();
									} else {
										instanceValue[j] = Double.parseDouble(currentStep.get(attribute).toString());
									}
								} else {
									instanceValue[j] = attributePossibleVals.get(attribute).indexOf(
											currentStep.get(attribute).toString());
								}
							} else if (!intervals.containsKey(attribute)) {
								instanceValue[j] = attributePossibleVals.get(attribute).indexOf(UNDEFINED_VALUE);
							}
							//							}
						}
					}

					//					if (timeAttributeMode != 0
					//							&& (dataAlignment.getStepTypes().get(i) != DataAwareStepTypes.MREAL || dataAlignment
					//									.getStepTypes().get(i) != DataAwareStepTypes.MINVI)) {
					//						instanceValue[instanceValue.length - 2] = getTimeClassificationValue(
					//								dataAlignment.getTraceName(), l, anchorTimestamp);
					//					} else if (timeAttributeMode != 0)
					//						instanceValue[instanceValue.length - 2] = getTimeClassificationValue(anchorTimestamp,
					//								latestRealTimestamp);
					if (timeAttributeMode != 0)
						instanceValue[instanceValue.length - 2] = timeClassificationValues.get(currentStep);

					int classValue;
					if (this.replayResultTemporal != null) {
						DashboardAlignment timeAlignment = findAlignmentWithLabel(dataAlignment.getTraceName(),
								this.replayResultTemporal.labelStepArray);
						classValue = getClassAttrValue(dataAlignment.getStepTypes().get(i), dataAlignment.getLogTrace()
								.get(i), dataAlignment.getProcessTrace().get(i), timeAlignment.getStepTypes().get(i));
					} else {
						classValue = getClassAttrValue(dataAlignment.getStepTypes().get(i), dataAlignment.getLogTrace()
								.get(i), dataAlignment.getProcessTrace().get(i));
					}

					instanceValue[instanceValue.length - 1] = classValue;
					Instance newInstance = new DenseInstance(1.0, instanceValue);
					for (int z = 0; z < includedAttributesList.size() - 2; z++) {
						if (currentStep.get(includedAttributesList.get(z).name()) == null
								&& !includedAttributesList.get(z).name().equals("Activity")) {
							newInstance.setMissing(includedAttributesList.get(z));
						}
					}
					instances.add(newInstance);
					//					anchorTimestamp = newAnchorTimestamp;
					//					try {
					//						instances.toString().startsWith("@relation StepInstances");
					//					} catch (Exception ex) {
					//						ex.printStackTrace();
					//					}
				} catch (Exception ex) {
					ex.printStackTrace();
				}
			}
		}
		return instances;
	}

	public XTrace findLogTrace(String traceName) {
		XTrace result = null;
		XConceptExtension ce = XConceptExtension.instance();
		for (int i = 0; i < this.log.size(); i++) {
			if (ce.extractName(this.log.get(i)).equals(traceName)) {
				result = this.log.get(i);
				break;
			}
		}

		return result;
	}

	//gets value in hours
	//TODO: allow user change unit
	public long getTimestampAttributeValue(String traceName, int eventIndex) {
		long result = -1;
		try {
			//			XAttributeTimestamp problemTimestampAttribute = (XAttributeTimestamp) findLogTrace(traceName)
			//					.get(eventIndex).getAttributes().get("time:timestamp");
			//			result = problemTimestampAttribute.getValue().getTime();
			result = XTimeExtension.instance().extractTimestamp(findLogTrace(traceName).get(eventIndex)).getTime();
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		return result;
	}

	//gets value in hours
	//TODO: allow user change unit
	public float getTimeClassificationValue(String traceName, int eventIndex, long anchorTimestamp) {
		//		float result = -1;
		//		if (anchorTimestamp == -1)
		//			return result;
		long problemTimestamp = getTimestampAttributeValue(traceName, eventIndex);
		//		if (problemTimestamp != -1) {
		//			result = (problemTimestamp - anchorTimestamp) / (60.0f * 60.0f * 1000.0f);
		//		}
		return getTimeClassificationValue(anchorTimestamp, problemTimestamp);
	}

	public float getTimeClassificationValue(long anchorTimestamp, long problemTimestamp) {
		float result = -1;
		if (anchorTimestamp == -1)
			return result;
		if (problemTimestamp != -1) {
			result = Math.abs((problemTimestamp - anchorTimestamp) / (60.0f * 60.0f * 1000.0f));
		}
		return result;
	}

	//	public void configureClassAttributeVals(int classificationType) {
	//		this.classAttributeValsList = new ArrayList<String>();
	//		this.classAttributeValsList.add("non-problematic");
	//		if (classificationType == 0)
	//			this.activityClassificationInstanceCount = 0;
	//		else if (classificationType == 1)
	//			this.assignmentClassificationInstanceCount = 0;
	//		for (Alignment dataAlignment : this.replayResultWithData.labelStepArray) {
	//			for (int i = 0; i < dataAlignment.getLogTrace().size(); i++) {
	//				ExecutionStep currentStep;
	//				String currentActivity;
	//				if (dataAlignment.getStepTypes().get(i) == DataAwareStepTypes.MREAL) {
	//					currentStep = dataAlignment.getProcessTrace().get(i);
	//					currentActivity = currentStep.getActivity();
	//				} else {
	//					currentStep = dataAlignment.getLogTrace().get(i);
	//					currentActivity = dashboard.getActivityName(currentStep.getActivity());
	//				}
	//				if (!includeStepInInstances(currentStep, currentActivity, classificationType))
	//					continue;
	//				
	//				if (classificationType == 0)
	//					this.activityClassificationInstanceCount++;
	//				else if (classificationType == 1)
	//					this.assignmentClassificationInstanceCount++;
	//				
	//				DataAwareStepTypes dataStepType = dataAlignment.getStepTypes().get(i);
	//				if (dataStepType == DataAwareStepTypes.MINVI)
	//					continue;
	//				else if (dataStepType == DataAwareStepTypes.L && !this.classAttributeValsList.contains("log-only")) {
	//					this.classAttributeValsList.add("log-only");
	//				} else if (dataStepType == DataAwareStepTypes.MREAL
	//						&& !this.classAttributeValsList.contains("model-only")) {
	//					this.classAttributeValsList.add("model-only");
	//				} else if (dataStepType == DataAwareStepTypes.LMNOGOOD) {
	//					//					this.dashboard.
	//					String problem = "incorrect ";
	//					for (String attribute : this.attributeIndex) {
	//						if (!this.dashboard.ExecutionStepsEqual(attribute, dataAlignment.getProcessTrace().get(i),
	//								dataAlignment.getLogTrace().get(i)))
	//							problem += attribute + ", ";
	//					}
	//					if (this.replayResultTemporal != null) {
	//						Alignment timeAlignment = findAlignmentWithLabel(dataAlignment.getTraceName(),
	//								this.replayResultTemporal.labelStepArray);
	//						if (timeAlignment.getStepTypes().get(i) == DataAwareStepTypes.LMNOGOOD)
	//							problem += "Time";
	//					}
	//					if (problem.endsWith(", "))
	//						problem = problem.substring(0, problem.length() - 2);
	//
	//					if (!this.classAttributeValsList.contains(problem))
	//						this.classAttributeValsList.add(problem);
	//				} else if (this.replayResultTemporal != null) {
	//					Alignment timeAlignment = findAlignmentWithLabel(dataAlignment.getTraceName(),
	//							this.replayResultTemporal.labelStepArray);
	//					if (timeAlignment.getStepTypes().get(i) == DataAwareStepTypes.LMNOGOOD
	//							&& !this.classAttributeValsList.contains("Incorrect time"))
	//						this.classAttributeValsList.add("incorrect Time");
	//				}
	//			}
	//		}
	//	}

	public void configureClassAttributeVals(int classificationType) {
		this.classAttributeValsList = new ArrayList<String>();
		this.classAttributeValsList.add("non-problematic");
		this.classAttributeValsList.add("problematic");
	}

	//	private int getClassAttrValue(DataAwareStepTypes dataStepType, ExecutionStep logStep, ExecutionStep processStep,
	//			DataAwareStepTypes timeStepType) {
	//		if (dataStepType == DataAwareStepTypes.L) {
	//			return this.classAttributeValsList.indexOf("log-only");
	//		} else if (dataStepType == DataAwareStepTypes.MREAL) {
	//			return this.classAttributeValsList.indexOf("model-only");
	//		} else if (dataStepType == DataAwareStepTypes.LMNOGOOD) {
	//			String problem = "Incorrect ";
	//			for (String attribute : this.attributeIndex) {
	//				if (!this.dashboard.ExecutionStepsEqual(attribute, processStep, logStep))
	//					problem += attribute + ", ";
	//			}
	//			if (timeStepType == DataAwareStepTypes.LMNOGOOD)
	//				problem += "Time";
	//			if (problem.endsWith(", "))
	//				problem = problem.substring(0, problem.length() - 2);
	//			return this.classAttributeValsList.indexOf(problem);
	//		} else {
	//			return this.classAttributeValsList.indexOf("non-problematic");
	//		}
	//	}

	private int getClassAttrValue(DataAwareStepTypes dataStepType, ExecutionStep logStep, ExecutionStep processStep,
			DataAwareStepTypes timeStepType) {
		if (this.problemSentenceObject.getViolationClass() != 3) {
			return getClassAttrValue(dataStepType, logStep, processStep);
		} else if (timeStepType == DataAwareStepTypes.LMNOGOOD && this.problemSentenceObject.getViolationClass() == 3) {
			return this.classAttributeValsList.indexOf("problematic");
		} else {
			return this.classAttributeValsList.indexOf("non-problematic");
		}
	}

	private int getClassAttrValue(DataAwareStepTypes dataStepType, ExecutionStep logStep, ExecutionStep processStep) {
		boolean isDecisionPoint = logStep.keySet().size() > 0 && processStep.keySet().size() > 0;
		if (dataStepType == DataAwareStepTypes.L && this.problemSentenceObject.getViolationClass() == 0) {
			return this.classAttributeValsList.indexOf("problematic");
		} else if (dataStepType == DataAwareStepTypes.MREAL && this.problemSentenceObject.getViolationClass() == 1) {
			return this.classAttributeValsList.indexOf("problematic");
		} else if (((dataStepType == DataAwareStepTypes.MINVI && isDecisionPoint) || dataStepType == DataAwareStepTypes.LMNOGOOD)
				&& this.problemSentenceObject.getViolationClass() == 2) {
			if (!this.dashboard.ExecutionStepsEqual(this.problemSentenceObject.getViolatedAttribute(), processStep,
					logStep))
				return this.classAttributeValsList.indexOf("problematic");
			else
				return this.classAttributeValsList.indexOf("non-problematic");
		} else {
			return this.classAttributeValsList.indexOf("non-problematic");
		}
	}

	//	private int getClassAttrValue(DataAwareStepTypes dataStepType, ExecutionStep logStep, ExecutionStep processStep) {
	//		if (dataStepType == DataAwareStepTypes.L) {
	//			return this.classAttributeValsList.indexOf("log-only");
	//		} else if (dataStepType == DataAwareStepTypes.MREAL) {
	//			return this.classAttributeValsList.indexOf("model-only");
	//		} else if (dataStepType == DataAwareStepTypes.LMNOGOOD) {
	//			String problem = "incorrect ";
	//			for (String attribute : this.attributeIndex) {
	//				if (!this.dashboard.ExecutionStepsEqual(attribute, processStep, logStep))
	//					problem += attribute + ", ";
	//			}
	//			if (problem.endsWith(", "))
	//				problem = problem.substring(0, problem.length() - 2);
	//			return this.classAttributeValsList.indexOf(problem);
	//		} else {
	//			return this.classAttributeValsList.indexOf("non-problematic");
	//		}
	//	}

	//	private boolean includeStepInInstances(ExecutionStep step, String activity, int classificationType) {
	//		if (classificationType == 0) {
	//			if (this.problemSentenceObject.getActivity() != null
	//					&& this.problemSentenceObject.getActivity().length() > 0
	//					&& this.problemSentenceObject.getActivity().equals(activity))
	//				return true;
	//			else
	//				return false;
	//		} else if (classificationType == 1) {
	//			if (this.problemSentenceObject.getAttribute() != null
	//					&& this.problemSentenceObject.getAttribute().length() > 0
	//					&& step.containsKey(this.problemSentenceObject.getAttribute())) {
	//				boolean intervalMatch = false;
	//				if (this.intervals.containsKey(this.problemSentenceObject.getAttribute())) {
	//					if (isInInterval(this.problemSentenceObject.getAttribute(),
	//							((Number) step.get(this.problemSentenceObject.getAttribute())).floatValue(),
	//							this.problemSentenceObject.getAttributeValue())) {
	//						return true;
	//					} else {
	//						return false;
	//					}
	//				} else if (step.get(this.problemSentenceObject.getAttribute()).toString()
	//						.equals(this.problemSentenceObject.getAttributeValue())) {
	//					return true;
	//				} else {
	//					return false;
	//				}
	//			} else {
	//				return false;
	//			}
	//		} else if (classificationType == 2) {
	//		}
	//		return false;
	//	}

	private boolean includeStepInInstances(ExecutionStep step, String activity, int classificationType) {
		if (classificationType == 0) {
			if (this.problemSentenceObject.getActivity() != null
					&& this.problemSentenceObject.getActivity().length() > 0
					&& this.problemSentenceObject.getActivity().equals(activity)
					|| this.problemSentenceObject.getActivity() == null)
				return true;
		} else if (classificationType == 1) {
			if (this.problemSentenceObject.getAttribute() == null
					|| this.problemSentenceObject.getAttribute().length() == 0) {
				return true;
			} else if (this.problemSentenceObject.getAttribute() != null
					&& this.problemSentenceObject.getAttribute().length() > 0
					&& step.containsKey(this.problemSentenceObject.getAttribute())) {
				if (this.intervals.containsKey(this.problemSentenceObject.getAttribute())) {
					if (isInInterval(this.problemSentenceObject.getAttribute(),
							((Number) step.get(this.problemSentenceObject.getAttribute())).floatValue(),
							this.problemSentenceObject.getAttributeValue())) {
						//						if (!step.get(this.problemSentenceObject.getAttribute()).toString().equals("0.0")) {
						//							int a = 5;
						//							int b = a + 5;
						//						}
						return true;
					} else {
						return false;
					}
				} else if (step.get(this.problemSentenceObject.getAttribute()) != null
						&& step.get(this.problemSentenceObject.getAttribute()).toString()
								.equals(this.problemSentenceObject.getAttributeValue())) {
					return true;
				} else {
					return false;
				}
			}
		} else if (classificationType == 2) {
			boolean activityMatch = false, assignmentMatch = false;
			if (this.problemSentenceObject.getActivity() != null
					&& this.problemSentenceObject.getActivity().length() > 0
					&& this.problemSentenceObject.getActivity().equals(activity)
					|| this.problemSentenceObject.getActivity() == null)
				activityMatch = true;

			if (this.problemSentenceObject.getAttribute() == null
					|| this.problemSentenceObject.getAttribute().length() == 0) {
				assignmentMatch = true;
			} else if (this.problemSentenceObject.getAttribute() != null
					&& this.problemSentenceObject.getAttribute().length() > 0
					&& step.containsKey(this.problemSentenceObject.getAttribute())) {
				if (this.intervals.containsKey(this.problemSentenceObject.getAttribute())) {
					if (isInInterval(this.problemSentenceObject.getAttribute(),
							((Number) step.get(this.problemSentenceObject.getAttribute())).floatValue(),
							this.problemSentenceObject.getAttributeValue())) {
						assignmentMatch = true;
					} else {
						assignmentMatch = false;
					}
				} else if (step.get(this.problemSentenceObject.getAttribute()) != null
						&& step.get(this.problemSentenceObject.getAttribute()).toString()
								.equals(this.problemSentenceObject.getAttributeValue())) {
					assignmentMatch = true;
				} else {
					assignmentMatch = false;
				}
			}
			return activityMatch && assignmentMatch;
		}
		return false;
	}

	// private boolean isStepTimeProblematic(ExecutionStep step,
	// DataAwareStepTypes timeStepType) {
	//
	// }

	private boolean isInInterval(String attribute, float value, String interval) {
		if (!intervals.containsKey(attribute))
			return false;
		boolean result = false;
		int intervalIndex = Integer.parseInt(interval);
		if (intervalIndex == 0)
			result = value <= intervals.get(attribute)[intervalIndex];
		else
			result = value <= intervals.get(attribute)[intervalIndex]
					&& value > intervals.get(attribute)[intervalIndex - 1];
		return result;
	}

	private DashboardAlignment findAlignmentWithLabel(String label, Collection<DashboardAlignment> labelStepArray) {
		DashboardAlignment result = null;
		for (DashboardAlignment alignment : labelStepArray)
			if (alignment.getTraceName().equalsIgnoreCase(label)) {
				result = alignment;
				break;
			}
		return result;
	}

	private J48 generateTree(Instances instances, String[] options) {
		J48 tree = new J48();
		try {
			tree.setOptions(options);
			tree.buildClassifier(instances);
			return tree;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public String getActivityGraph() {
		try {
			if (this.activityTree != null)
				return this.activityTree.graph();
			else
				return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public List<String> getActivityClassValues() {
		return this.activityClassificationClassValues;
	}

	public String getAssignmentGraph() {
		try {
			if (this.assignmentTree != null)
				return this.assignmentTree.graph();
			else
				return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public List<String> getAssignmentClassValues() {
		return this.assignmentClassificationClassValues;
	}

	public String getIntersectionGraph() {
		try {
			if (this.intersectionTree != null)
				return this.intersectionTree.graph();
			else
				return null;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public List<String> getIntersectionClassValues() {
		return this.intersectionClassificationClassValues;
	}

	public void selectSignificantAttributes() {
		if (this.problemSentenceObject.getAttribute() != null && this.problemSentenceObject.getAttribute().length() > 0) {
			this.assignmentClassificationAttributes = new ArrayList<String>();
		}
		if (this.problemSentenceObject.getActivity() != null && this.problemSentenceObject.getActivity().length() > 0)
			this.activityClassificationAttributes = new ArrayList<String>();

		if (this.activityClassificationAttributes != null && this.activityClassificationAttributes != null)
			this.intersectionClassificationAttributes = new ArrayList<String>();

		for (String attribute : this.dashboard.getAttributeImportanceMap().keySet()) {
			if (this.assignmentClassificationAttributes != null
					&& !this.problemSentenceObject.getAttribute().equals(attribute)) {
				this.assignmentClassificationAttributes.add(attribute);
				if (this.intersectionClassificationAttributes != null)
					this.intersectionClassificationAttributes.add(attribute);
			}
			if (this.activityClassificationAttributes != null)
				this.activityClassificationAttributes.add(attribute);
		}
	}
	
	public static DefaultMutableTreeNode buildTree(String graph) {
		graph = graph.substring(graph.indexOf("{") + 1, graph.indexOf("}"));
		DefaultMutableTreeNode result = null;
		Map<String, DefaultMutableTreeNode> IDNodeMap = new HashMap<String, DefaultMutableTreeNode>();
		Pattern pattern = Pattern.compile("(N[0-9]+(->N[0-9]+){0,1}) *\\u005B([^\\u005D]+)\\u005D");
		Matcher matcher = pattern.matcher(graph);
		boolean isRoot = true;
		while (matcher.find()) {
			if (!isRoot && matcher.group(3).contains("problematic (0.0)"))
				continue;
			isRoot = false;
			DefaultMutableTreeNode newNode = new DefaultMutableTreeNode(extractLabel(matcher.group(3)));
			IDNodeMap.put(matcher.group(1), newNode);
			if (IDNodeMap.size() == 1) {
				result = newNode;
			}
		}
		pattern = Pattern.compile("(N[0-9]+)->(N[0-9]+)");
		for (Entry<String, DefaultMutableTreeNode> nodeEntry : IDNodeMap.entrySet()) {
			matcher = pattern.matcher(nodeEntry.getKey());
			if (matcher.find()) {
				IDNodeMap.get(matcher.group(1)).add(nodeEntry.getValue());
				DefaultMutableTreeNode newChild = IDNodeMap.get(matcher.group(2));
				if (newChild != null)
					nodeEntry.getValue().add(newChild);
			}
		}

		pruneTree(result);
		return result;
	}

	public static void pruneTree(DefaultMutableTreeNode rootNode) {
		for (int i = 0; i < rootNode.getChildCount(); i++) {
			pruneTree((DefaultMutableTreeNode) rootNode.getChildAt(i));
			if (rootNode.getChildAt(i).isLeaf() && !rootNode.getChildAt(i).toString().contains("problematic")) {
				rootNode.remove(i);
				i--;
			}
		}
	}

	public static List<List<String>> findPreconditions(String problem, DefaultMutableTreeNode treeRoot, int timeMode, Instances instances) {
		List<List<String>> result = null;
		List<TreeNode> leafList = new ArrayList<TreeNode>();
		Pattern pattern = Pattern.compile("^" + problem + " *\\(([0-9]+\\.[0-9]+)(/([0-9]+\\.[0-9]+))?\\)");
		findLeafs(treeRoot, leafList, pattern);
		for (TreeNode leaf : leafList) {
			Matcher matcher = pattern.matcher(leaf.toString());
			String confidence = "";
			String frequency = "";
			try {
				if (matcher.find()) {
					if (matcher.group(3) == null) {
						confidence = "100%";
						frequency = matcher.group(1).replace(".0", "");
					} else {
						float frequencyFloat = Float.parseFloat(matcher.group(1)) - Float.parseFloat(matcher.group(3));
						float confidenceFloat = frequencyFloat / Float.parseFloat(matcher.group(1));
						confidenceFloat = Math.round(confidenceFloat * 1000.0f) / 10.0f;
						confidence += confidenceFloat + "%";

						frequency += ((int) frequencyFloat);
					}
				}
			} catch (Exception ex) {
				ex.printStackTrace();
			}

			Map<String, float[]> attributeIntervalsMap = new HashMap<String, float[]>();
			List<String> precondition = new ArrayList<String>();
			TreeNode parent = leaf.getParent();
			TreeNode oldParent = parent;
			while (parent != null) {
				if (parent.toString().startsWith("<=") || parent.toString().startsWith(">=")
						|| parent.toString().startsWith(">") || parent.toString().startsWith("<")) {
					oldParent = parent;
					parent = parent.getParent();
				} else if (parent.toString().startsWith("=")) {
					oldParent = parent;
					parent = parent.getParent();
				} else {
					if (oldParent.toString().startsWith("<=") || oldParent.toString().startsWith(">=")
							|| oldParent.toString().startsWith(">") || oldParent.toString().startsWith("<")) {
						float value = Float.parseFloat(oldParent.toString().replace("<=", "").replace(">=", "")
								.replace("<", "").replace(">", "").trim());
						float[] interval = attributeIntervalsMap.get(parent.toString());
						if (interval == null) {
							interval = new float[4];
							interval[0] = -Float.MAX_VALUE;
							interval[1] = Float.MAX_VALUE;
							interval[2] = 1.0f;
							interval[3] = 1.0f;
							attributeIntervalsMap.put(parent.toString(), interval);
						}
						if (oldParent.toString().startsWith("<=") && interval[1] > value) {
							interval[1] = value;
							interval[3] = 1.0f;
						} else if (oldParent.toString().startsWith(">=") && interval[0] < value) {
							interval[0] = value;
							interval[2] = 1.0f;
						} else if (oldParent.toString().startsWith("<") && interval[0] >= value) {
							interval[1] = value;
							interval[3] = 0.0f;
						} else if (oldParent.toString().startsWith(">") && interval[0] <= value) {
							interval[0] = value;
							interval[2] = 0.0f;
						}
						parent = parent.getParent();
					} else if (oldParent.toString().startsWith("=")) {
						//						precondition += parent.toString() + oldParent.toString() + " AND ";
						precondition.add(parent.toString() + oldParent.toString());
						parent = parent.getParent();
					}
				}
			}
			for (Entry<String, float[]> attributeEntry : attributeIntervalsMap.entrySet()) {
				String lowerBound = "", upperBound = "";
				if (attributeEntry.getKey().equals("timeClass")) {
					if (attributeEntry.getValue()[1] == -1) {
						if (timeMode == 1) {
							precondition.add("Activity is first in the case");
						} else if (timeMode == 2) {
							precondition.add("Activity is first in the case");
						} else if (timeMode == 3) {
							precondition.add("Activity occurs before the selected one");
						}
					} else {
						if (attributeEntry.getValue()[0] == -1)
							attributeEntry.getValue()[0] = 0;

						if (attributeEntry.getValue()[0] > -Float.MAX_VALUE) {
							float newMinimum = InstanceClassifier.findNewMinimum(instances, attributeEntry.getValue()[0], "timeClass");
							precondition.add(timePrefixes[timeMode - 1] + " is above " + newMinimum
									+ " hours");
						}
						if (attributeEntry.getValue()[1] == 0) {
							if (timeMode == 1) {
								precondition.add("Activity is first in the case");
							} else if (timeMode == 2) {
								precondition.add("Activity occurs simultaneously with the previous one");
							} else if (timeMode == 3) {
								precondition.add("Activity occurs simultaneously with the selected one");
							}
						} else if (attributeEntry.getValue()[1] < Float.MAX_VALUE) {
							precondition.add(timePrefixes[timeMode - 1] + " is below " + +attributeEntry.getValue()[1]
									+ " hours");
						}
					}
				} else {
					if (attributeEntry.getValue()[0] > -Float.MAX_VALUE) {
						if (attributeEntry.getValue()[2] == 0)
							lowerBound += " > " + attributeEntry.getValue()[0];
						else if (attributeEntry.getValue()[2] == 1)
							lowerBound += " >= " + attributeEntry.getValue()[0];
						precondition.add(attributeEntry.getKey() + lowerBound);
					}
					if (attributeEntry.getValue()[1] < Float.MAX_VALUE) {
						if (attributeEntry.getValue()[3] == 0)
							upperBound += " < " + attributeEntry.getValue()[1];
						else if (attributeEntry.getValue()[3] == 1)
							upperBound += " <= " + attributeEntry.getValue()[1];
						precondition.add(attributeEntry.getKey() + upperBound);
					}
				}
			}
			precondition.add("Frequency: " + frequency + "; confidence: " + confidence);
			if (result == null)
				result = new ArrayList<List<String>>();
			result.add(precondition);
		}
		return result;
	}

	public static float findNewMinimum(Instances instances, float origMinValue, String attributeName) {
		float result = Float.MAX_VALUE;
		Attribute attribute = null;
		while (instances.enumerateAttributes().hasMoreElements()) {
			Attribute attr = (Attribute) instances.enumerateAttributes().nextElement();
			if (attr.name().equals(attributeName)) {
				attribute = attr;
				break;
			}
		}
		if (attribute == null)
			return origMinValue;

		for (int i = 0; i < instances.size(); i++) {
			if ((float) instances.get(i).value(attribute) > origMinValue
					&& (float) instances.get(i).value(attribute) < result)
				result = (float) instances.get(i).value(attribute);
		}

		return result;
	}

	public static void findLeafs(TreeNode root, List<TreeNode> leafList, Pattern pattern) {
		Enumeration children = root.children();

		while (children.hasMoreElements()) {
			TreeNode child = (TreeNode) children.nextElement();
			if (child.isLeaf()) {
				Matcher matcher = pattern.matcher(child.toString());
				if (matcher.find())
					leafList.add(child);
			} else
				findLeafs(child, leafList, pattern);
		}
	}

	public static String extractLabel(String labelString) {
		Pattern pattern = Pattern.compile("label=\"([^\"]+)\"");
		Matcher matcher = pattern.matcher(labelString);
		if (matcher.find())
			return matcher.group(1);
		else
			return "";
	}

	public List<String> getActivityClassificationAttributes() {
		return this.activityClassificationAttributes;
	}

	public List<String> getAssignmentClassificationAttributes() {
		return this.assignmentClassificationAttributes;
	}

	public List<String> getIntersectionClassificationAttributes() {
		return this.intersectionClassificationAttributes;
	}

	public void setActivityTimeAttributeMode(int mode) {
		this.activityTimeAttributeMode = mode;
	}

	public int getActivityTimeAttributeMode() {
		return this.activityTimeAttributeMode;
	}

	public void setAssignmentTimeAttributeMode(int mode) {
		this.assignmentTimeAttributeMode = mode;
	}

	public int getAssignmentTimeAttributeMode() {
		return this.assignmentTimeAttributeMode;
	}

	public void setIntersectionTimeAttributeMode(int mode) {
		this.intersectionTimeAttributeMode = mode;
	}

	public int getIntersectionTimeAttributeMode() {
		return this.intersectionTimeAttributeMode;
	}

	public void setActivityAnchor(String activityAnchor) {
		this.activityAnchor = activityAnchor;
	}

	public String getActivityAnchor() {
		return this.activityAnchor;
	}

	public void setAssignmentAnchor(String assignmentAnchor) {
		this.assignmentAnchor = assignmentAnchor;
	}

	public String getAssignmentAnchor() {
		return this.assignmentAnchor;
	}

	public void setIntersectionAnchor(String intersectionAnchor) {
		this.intersectionAnchor = intersectionAnchor;
	}

	public String getIntersectionAnchor() {
		return this.intersectionAnchor;
	}

	public Instances getActivityInstances() {
		return this.activityInstances;
	}

	public Instances getAssignmentInstances() {
		return this.assignmentInstances;
	}

	public Instances getIntersectionInstances() {
		return this.intersectionInstances;
	}
}